/**
 * Check if images was loaded.
 *
 * @return {Number}
 * @see images_monitor.js
 *
 * Returns 0 if images was not loaded.
 *
 * Returns anything above 0 if images was loaded.
 */

return window.__selenium_was_equal;
